
#include <malloc.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>
#include <errno.h>

/* ucd/net-snmp includes */
#define HAVE_STDARG_H 1
#include <asn1.h>
#include <snmp_api.h>
#include <snmp_impl.h>
#include <snmp_client.h>
#include <mib.h>
#include <snmp.h>
#include <default_store.h>

#undef LIST_DEBUG
#undef DEBUG

#ifdef DEBUG
#define DBGPRINT 	printf
#else
#define DBGPRINT( a... )
#endif

#include "list.h"
#include "asyncsnmp.h"

#define INBUF_SIZE			8192

#define DEFAULT_SESSIONMAX		768
#define DEFAULT_COMMUNITY		"public"
#define DEFAULT_MAXPENDING		5

int schedule_request(struct async_session *asess, struct async_request *areq);

LIST_HEAD(struct async_session)		active_sessions;
LIST_HEAD(struct async_session)		closeing_sessions;
LIST_HEAD(struct async_session)		pending_sessions;

int					sessactive=0,
					validin=0,
					request=0,
					response=0;

char					*inbuf,	
					*outbuf;

void tvsub(struct timeval *tv, struct timeval *a, struct timeval *b) {
	tv->tv_sec = a->tv_sec - b->tv_sec;
	tv->tv_usec = a->tv_usec - b->tv_usec;
	if (tv->tv_usec < 0) {
		tv->tv_sec--;
		tv->tv_usec += 10000000;
	}
}

void free_request(struct async_request *ar) {

	if (ar->ureqid)
		free(ar->ureqid);

	free(ar);
}


void process_reply(struct async_session *asess, struct async_request *areq, 
		struct snmp_pdu *pdu, struct snmp_session *sp, int operation) {

	struct variable_list 	*vp;
	struct timeval		tv;
	char			buf[256];
	int			keepreq=0,
				abortwalk=0;
	struct snmp_pdu		*sreq;

	/* Get current time (the response arrived) */
	gettimeofday(&tv, NULL);

	switch(operation) {
		case RECEIVED_MESSAGE:
			response++;

			/* SNMP error (nosuchname etc) */
			if (pdu->errstat != SNMP_ERR_NOERROR) {

				fprintf(stdout, "%lu.%03lu %s error %s\n",
					tv.tv_sec, tv.tv_usec / 1000,
					areq->ureqid, snmp_errstring(pdu->errstat));
				
				break;
			} 

			/* Print our variables */
			for(vp=pdu->variables;vp;vp=vp->next_variable) {
				/* If we got an oid shorter than we requested
				 * or we got a completely bogus answer */
				if (vp->name_length < areq->oidlen ||
					(memcmp(areq->oid, vp->name, areq->oidlen * sizeof(oid)) != 0)) {

					if (areq->reqtype == SNMP_MSG_GET) {
						fprintf(stdout, "%lu.%03lu %s error got bogus result var\n",
								tv.tv_sec, tv.tv_usec / 1000, areq->ureqid);
					} else {
						 fprintf(stdout, "%lu.%03lu %s success EOT\n",
								tv.tv_sec, tv.tv_usec / 1000, areq->ureqid);
						abortwalk=1;
					}

					break;
				} else {
					/* FIXME - Replace with sprint_oid and sprint_value 
					 * as we dont want the oid to be decoded - Too much overhead */
					sprint_variable(buf, vp->name, vp->name_length, vp);

					fprintf(stdout, "%lu.%03lu %s success %s\n",
						tv.tv_sec, tv.tv_usec / 1000, areq->ureqid, buf);
				}
			}

			/* Schedule next request */
			if (areq->reqtype == SNMP_MSG_GETNEXT &&
				!abortwalk) {

				sreq = snmp_pdu_create(areq->reqtype);		
				snmp_add_null_var(sreq, pdu->variables->name,
								pdu->variables->name_length);

				/* Copy request ID */
				areq->reqid=sreq->reqid;

				if(snmp_send(asess->sess, sreq) == 0) {
					snmp_perror("snmp_send");
					snmp_free_pdu(sreq);
				} else 
					keepreq=1;

				request++;
			}

			break;
		case SEND_FAILED:
			fprintf(stdout, "%lu.%03lu %s error send failed in async_callback\n",
					tv.tv_sec, tv.tv_usec / 1000, areq->ureqid);
			break;
		case TIMED_OUT:
			fprintf(stdout, "%lu.%03lu %s error timed out in async_callback\n",
					tv.tv_sec, tv.tv_usec / 1000, areq->ureqid);
			break;
		default:
			fprintf(stdout, "%lu.%03lu %s error code %d in async_callback unknown\n",
					tv.tv_sec, tv.tv_usec / 1000, areq->ureqid, operation);
	}

	if (!keepreq) {
		LIST_DELETE(areq);
		free_request(areq);

		asess->reqpending--;
		asess->reqtotal--;
	}
}


int real_async_callback(int operation, struct snmp_session *sp,
		int reqid, struct snmp_pdu *pdu, void *magic) {
	struct async_session	*asess = (struct async_session *) magic;
	struct async_request	*areq=NULL,
				*ar;

	DBGPRINT("%s %d\n", __FUNCTION__, reqid);

	/*
	 * Traverse the request list until we find the responses request
	 * or the end of the list
	 */
	LIST_WALK(&asess->req, ar) {
		if (ar->reqid == reqid) {
			areq=ar;
			break;
		}
	}


	/* Process reply and probably send out new GETNEXT */
	if (areq)
		process_reply(asess, areq, pdu, sp, operation);

	/* If we dont have any requests pending - Put session on closing
	 * session list */
	if (LIST_EMPTY(&asess->req) && asess->finished) {
		LIST_DELETE(asess);
		LIST_APPEND(&closeing_sessions, asess);
	} else {
		/* Try to send out new requests if this session
		 * is below the limit */
		LIST_WALK(&asess->req, areq) {

			if(asess->reqpending >= asess->maxpending ||
				asess->reqtotal < asess->reqpending)
					break;

			if (!areq->reqid) 
				schedule_request(asess, areq);
		}
	}

	return(1);
}		

int async_callback(int operation, struct snmp_session *sp,
			int reqid, struct snmp_pdu *pdu, void *magic) {

	return real_async_callback(operation, sp, reqid, pdu, magic);
}

struct snmp_session *open_session(struct async_session *as) {
	struct snmp_session	ss;

	DBGPRINT("Opening new session for %s\n", as->host);

	snmp_sess_init(&ss);

	ss.version = as->version;
	ss.peername = as->host;
	ss.community = as->community;
	ss.community_len = strlen(as->community);
	ss.callback = async_callback;
	ss.callback_magic = as;

	if (!(as->sess = snmp_open(&ss))) {
		snmp_perror("snmp_open");
		return(NULL);
	}	

	return(as->sess);
}

int schedule_request(struct async_session *as, struct async_request *ar) {
	struct snmp_pdu		*sreq;

	/* Create Request Packet */
	sreq = snmp_pdu_create(ar->reqtype);		
	snmp_add_null_var(sreq, ar->oid, ar->oidlen);

	/* Copy request ID */
	ar->reqid=sreq->reqid;	

	DBGPRINT("scheduling new request for %s id %s\n", as->host, ar->ureqid);

	/* Send packet */
	if (snmp_send(as->sess, sreq)) {
		as->reqpending++;
	} else {
		snmp_perror("snmp_send");

		/* Free request */
		snmp_free_pdu(sreq);

		LIST_DELETE(ar);

		free_request(ar);

		return(1);
	}
	request++;

	return(0);
}

#define PL_MAX_ARGS	20
struct parsed_line {
	int	argc;	
	char	*argv[PL_MAX_ARGS];
};

struct parsed_line *parse_line(char *line) {
	static struct parsed_line	pl;
	int				i,j;

	pl.argc=0;

	j=strlen(line);
	for(i=0;i<j;i++) {
		if(line[i] == '"') {
			for(i++;i<j && line[i] != '"';i++);
			i++;
		}

		for(;i<j && line[i] == ' ';line[i++]=0x0);

		pl.argv[pl.argc++]=&line[i];

		for(;i<j && line[i] != ' '
			&& line[i] != '"';i++);

		i--;
	}

	return(&pl);
}

/*
 * session host=127.0.0.1 community=public version=2 timeout=5
 * id get|walk oid
 * end
 */

struct async_session *create_session(struct parsed_line *pl) {
	int 	i;
	char	*eq;
	struct	async_session	*as;

	assert((as=(void *) malloc(sizeof(struct async_session))));
	memset(as, 0, sizeof(struct async_session));

	as->maxpending=DEFAULT_MAXPENDING;
	as->version=SNMP_VERSION_1;

#define SESSION_PARM_HOST	"host"
#define SESSION_PARM_COMMUNITY	"community"
#define SESSION_PARM_VERSION	"version"

	for(i=1;i<pl->argc;i++) {
		if (strncmp(pl->argv[i], SESSION_PARM_HOST, strlen(SESSION_PARM_HOST)) == 0) {
			if ((eq=strstr(pl->argv[i], "="))) {
				as->host=strdup(&eq[1]);
			}
		} else if (strncmp(pl->argv[i], SESSION_PARM_COMMUNITY,
					strlen(SESSION_PARM_COMMUNITY)) == 0) {
			if ((eq=strstr(pl->argv[i], "="))) {
				as->community=strdup(&eq[1]);
			}
		} else if (strncmp(pl->argv[i], SESSION_PARM_VERSION,
					strlen(SESSION_PARM_VERSION)) == 0) {
			if ((eq=strstr(pl->argv[i], "="))) {

				/* Skip space */
				eq++;

				if (strcmp(eq,"1") == 0)
					as->version=SNMP_VERSION_1;
				else if (strcmp(eq,"2") == 0)
					as->version=SNMP_VERSION_2c;
				else
					fprintf(stderr, "Unknown snmp version %s\n", eq);
			}
		} else {
			fprintf(stderr, "Syntax error\n");
		}
	}

	if (!as->host) {
		fprintf(stderr, "No hostname/ip address\n");
		if (as->community)
			free(as->community);
		free(as);
		return(NULL);
	}

	if (!as->community) 
		as->community=strdup(DEFAULT_COMMUNITY);

	LIST_INIT(&as->req);

	if (sessactive < DEFAULT_SESSIONMAX) {
		if (!open_session(as)) {

			free(as->community);
			free(as->host);
			free(as);
			
			return(NULL);
		} else {
			LIST_APPEND(&active_sessions, as);
			sessactive++;
		}
	} else {
		LIST_APPEND(&pending_sessions, as);
	}

	return(as);
}

void add_request(struct async_session *as, struct parsed_line *pl) {
	struct async_request		*ar;

	assert((ar=(void *)malloc(sizeof(struct async_request))));
	memset(ar, 0, sizeof(struct async_request));

#define REQTYPE_GET	"get"
#define REQTYPE_WALK	"walk"

	if (strcmp(pl->argv[1], REQTYPE_GET) == 0) {
		ar->reqtype=SNMP_MSG_GET;
	} else if (strcmp(pl->argv[1], REQTYPE_WALK) == 0) {
		ar->reqtype=SNMP_MSG_GETNEXT;
	} else {
		fprintf(stderr, "Unknown request type %s\n", pl->argv[1]);
		free(ar);
		return;
	}

	assert((ar->ureqid=strdup(pl->argv[0])));

	ar->oidlen=MAX_OID_LEN;

	if (snmp_parse_oid(pl->argv[2], ar->oid, &ar->oidlen) == NULL) {
		snmp_perror(pl->argv[2]);

		free(ar->ureqid);
		free(ar);

		return;
	}

	LIST_APPEND(&as->req, ar);

	as->reqtotal++;

	/* If we have an open session and request slots */
	if (as->sess && as->reqpending < as->maxpending)
		schedule_request(as, ar);
		
}

void process_line(char *line) {
	static struct async_session	*as;
	struct parsed_line		*pl;

	pl=parse_line(line);

	if (pl->argc >= 1) {

		if (strcmp(pl->argv[0], "session") == 0) {
			if (as) {
				fprintf(stderr, "Session was not properly ended\n");
				as->finished=1;
			}
			as=create_session(pl);
			return;
		} else if (strcmp(pl->argv[0], "end") == 0) {
			DBGPRINT("marking session %s as finished\n", as->host);
			as->finished=1;
			/* Not in session context */
			as=NULL;
			return;
		}

		if (!as) {
			fprintf(stderr, "Not in session context\n");
			return;
		}

		if (pl->argc != 3) {
			fprintf(stderr, "Wrong number of args in session context\n");
			return;
		}

		add_request(as, pl);
	}
}

/* We get called if fd is ready to read */
int readfile() {
	ssize_t		rbytes,
			toread;
	int		eof=0,
			sol,
			i;

	toread=INBUF_SIZE - validin;

	rbytes=read(fileno(stdin), &inbuf[validin], toread);

	/* If we have an error or read nothing we are boned aka EOF */
	if ((rbytes == -1 && errno != 0) || rbytes < toread)
		eof=1;

	/* update valid bytes in inbuffer */
	validin+=rbytes;

	DBGPRINT("readfile tried to read %d of which %d succeeded giving %d in buffer\n",
			toread, rbytes, validin);

	sol=0;
	/* We still have valid input */
	while(1) {
		for(i=sol;i<=validin && inbuf[i] != 0xa;i++);

		if (inbuf[i]==0xa || (eof && i==validin)) {
			inbuf[i]=0x0;

			DBGPRINT("calling process_line for %d/%d %s\n", i, sol, &inbuf[sol]);

			process_line(&inbuf[sol]);

			sol=i+1;
		} else 
			break;
	}

	/* Still input to process ? */
	if (sol >= validin) {
		validin=0;
	} else {

		/* Calculate leftover bytes in buffer */
		i=validin-sol;

		/* Move half line in buffer to front */
		memmove(inbuf, &inbuf[sol], i);
		validin=i;
	}

	return(eof);
}

int isfifo(int fd) {
	struct stat	st;

	fstat(fd, &st);

	return S_ISFIFO(st.st_mode);
}

void session_housekeeping() {

	while(!LIST_EMPTY(&closeing_sessions)) {
		struct async_session	*stmp;

		DBGPRINT("Closing session for host %s\n", closeing_sessions.next->sess->peername);

		snmp_close(closeing_sessions.next->sess);
		sessactive--;

		stmp=closeing_sessions.next;
		LIST_DELETE(stmp);

		free(stmp);
	}

	/* May we open more sessions */
	while(sessactive < DEFAULT_SESSIONMAX && !LIST_EMPTY(&pending_sessions)) {
		struct async_session	*asess;

		asess=pending_sessions.next;	

		/* Delete session from pending sessions list */
		LIST_DELETE(asess);

		/* Open session */
		if (!open_session(asess)) {
			/* It failed - so free all requests */
			while(!LIST_EMPTY(&asess->req)) {
				struct async_request	*areq;
			
				areq=asess->req.next;

				LIST_DELETE(areq);

				free_request(areq);
			}

			free(asess);

			/* We failed once - quit for now */
			break;
		} else {
			/* We succeeded in opening the session */
			struct async_request	*areq;

			/* Put on active sessions list */
			LIST_APPEND(&active_sessions, asess);

			/* Walk request list */
			LIST_WALK(&asess->req, areq) {
				if (asess->reqpending >= asess->maxpending)
					break;
				if (schedule_request(asess, areq))
					break;
			}
			sessactive++;
		}

	}
}

void process_input() {
	int			fds=0,
				eof=0,
				block,
				rounds=0,
				stdinflags;
	fd_set			fdset;	
	struct timeval		timeout,
				begin, end, runtime;


	/* Get stdin's flags */
	stdinflags = fcntl(fileno(stdin), F_GETFL, 0);

	gettimeofday(&begin, NULL);

	/* FIXME This is strange but works 
	 *
	 * Setting O_NONBLOCK on a fifo results in only 8192 bytes beeing returned
	 *
	 */
	if (!isfifo(fileno(stdin))) {
		DBGPRINT("Setting stdin(%d) to O_NONBLOCK\n", fileno(stdin));
		/* Set stdin non-blocking */
		fcntl(fileno(stdin), F_SETFL, stdinflags | O_NONBLOCK); 
	}

	/* Run until we have an EOF or if we still have sessions */
	while(!eof || sessactive > 0) {

		FD_ZERO(&fdset);

		block=1;

		snmp_select_info(&fds, &fdset, &timeout, &block);

		/* We dont need to try reading input when we reached EOF already */
		if (!eof) {
			/* FIXME We could only try stdin if we are low on processing */
			FD_SET(fileno(stdin), &fdset);

			/* If we are stdin only we need to set the maxfd accordingly */
			if (fds < fileno(stdin)+1)
				fds=fileno(stdin)+1;
		}

		/* Wait for the sessions and input */
		if ((fds=select(fds, &fdset, NULL, NULL, &timeout))) {

			DBGPRINT("Select returned %d pending fds\n", fds);
	
			/* Do we have input ? */
			if (FD_ISSET(fileno(stdin), &fdset)) {
				DBGPRINT("select meant stdin to be readable\n");
				eof=readfile();
			}

			snmp_read(&fdset);
			rounds++;
		} else {
			DBGPRINT("snmp_timeout called - real timeout\n");
			snmp_timeout();
			rounds=0;
			DBGPRINT("%d active session(s)\n", sessactive);
		}

		if (rounds > 30) {
			DBGPRINT("snmp_timeout called - virtual timeout\n");
			snmp_timeout();
			rounds=0;
		}

		/* Remove closing and queue pending sessions */
		session_housekeeping();
	}

	/* Restore flags in stdin */
	fcntl(fileno(stdin), F_SETFL, stdinflags);

	gettimeofday(&end, NULL);
	tvsub(&runtime, &end, &begin);

	fprintf(stdout, "-- time: %4lu.%03lu requests: %8d responses: %8d rps: %6lu\n",
		runtime.tv_sec, runtime.tv_usec / 1000,
		request, response,
		(response * 1000) / (runtime.tv_sec * 1000 + runtime.tv_usec / 1000));
}

int main(int argc, char **argv) {

	init_snmp("asyncsnmp");

	/* ds_toggle_boolean(DS_LIBRARY_ID, DS_LIB_PRINT_NUMERIC_OIDS);  */

	LIST_INIT(&active_sessions);
	LIST_INIT(&closeing_sessions);
	LIST_INIT(&pending_sessions);

	/* Allocate buffer + trailing zero */
	assert(inbuf=(char *) malloc(INBUF_SIZE+1));

	process_input();

	return(0);
}
